# @author Udo Schneider <Udo.Schneider@homeaddress.de>

module DeepSecurity

  # Represents an Anti-Malware event
  class AntiMalwareEvent < TransportObject

    attr_integer_accessor :anti_malware_config_id,
                          'The ID of the Anti-Malware configuration this event corresponds to'
    attr_integer_accessor :anti_malware_event_id,
                          'The ID of the event'
    attr_datetime_accessor :end_time,
                           'Endtime of this event if it was repeated multiple times (not currently used)'
    attr_integer_accessor :error_code,
                          'The VSAPI error code indicates the reason of the actions of failure'
    attr_integer_accessor :host_id,
                          'The host ID this event corresponds to'
    attr_string_accessor :infected_file_path,
                         'The infected file full path'
    attr_string_accessor :infection_source,
                         'The source computer of the infection'
    attr_datetime_accessor :log_date,
                           'The time this event occured'
    attr_string_accessor :malware_name,
                         'The name of the malware'
    attr_enum_accessor :malware_type,
                       EnumMalwareType,
                       'The type of the malware'
    attr_integer_accessor :protocol,
                          'The protocols: Local Files(0), Network shared folder(1), etc. However currently the Agent only support local files'
    attr_integer_accessor :quarantine_record_id,
                          'The ID of the quarantined file, if a file was quarantined as a result of this event.'
    attr_integer_accessor :scan_result_action1,
                          'The first action performed'
    attr_integer_accessor :scan_result_action2,
                          'The second action performed'
    attr_enum_accessor :scan_type,
                       EnumAntiMalwareScanType,
                       'Type of scan this event was captured under'
    array_object_accessor :spyware_items,
                          AntiMalwareSpywareItem,
                          'An array of spyware items associated with this event'
    attr_datetime_accessor :start_time,
                           'Starttime of this event if it was repeated multiple times (not currently used)'
    attr_string_accessor :tags,
                         'Any tags associated with this event'
    attr_integer_accessor :scan_action1
    attr_integer_accessor :scan_action2
    attr_string_accessor :summary_scan_result

    hint_object_accessor :host,
                         Host,
                         'The host this event corresponds to'


    # cache_by_aspect :anti_malware_event_id

    # @!group High-Level SOAP Wrapper

    def host
      manager.host(host_id)
    end

    # @!endgroup

  end

  class Manager

    # @!group High-Level SOAP Wrapper

    # Return all AntiMalware events matching the filter
    # @param time_filter [TimeFilter]
    # @param host_filter [HostFilter]
    # @param event_id_filter [IDFilter]
    # @return [Array<AntiMalwareEvent>]
    def anti_malware_events_by_time_host_event(time_filter, host_filter, event_id_filter)
      interface.antiMalwareEventRetrieve(time_filter, host_filter, event_id_filter)
    end

    # @!endgroup


  end

  class SOAPInterface

    # @!group Low-Level SOAP Wrapper

    # Retrieves the AntiMalware events specified by the time and host filter.
    #
    # SYNTAX
    #   public AntiMalwareEventListTransport antiMalwareEventRetrieve(TimeFilterTransport timeFilter HostFilterTransport hostFilter, IDFilterTransport eventIdFilter, String sID)
    #
    # PARAMETERS
    #   timeFilter Restricts the retrieved events by time.
    #   hostFilter Restricts the retrieved events by host, group, or security profile.
    #   eventIdFilter Restricts the retrieved events by event id.
    #   sID Authentication session identifier ID.
    #
    # RETURNS
    #   AntiMalwareEventListTransport object.
    def antiMalwareEventRetrieve(timeFilter, hostFilter, eventIdFilter, sID = manager.sID)
      request_array(:anti_malware_event_retrieve, AntiMalwareEvent, :anti_malware_events,
                    :timeFilter => timeFilter.to_savon,
                    :hostFilter => hostFilter.to_savon,
                    :eventIdFilter => eventIdFilter.to_savon,
                    :sID => sID)
    end

    # @!endgroup

  end

end
